//
//  OpenGL widget
//
#include "ex20opengl.h"
#include <QtOpenGL>
#include <QMessageBox>
#include <QGLFunctions>
#include <math.h>

//
//  Advance time one time step for star k
//
void Ex20opengl::Move(int k)
{
   int k0 = k+src;
   int k1 = k+dst;
   float dt = 1e-3;
   int i;
   //  Calculate force components
   double a=0;
   double b=0;
   double c=0;
   for (i=src;i<src+N;i++)
   {
      double dx = stars[i].x-stars[k0].x;
      double dy = stars[i].y-stars[k0].y;
      double dz = stars[i].z-stars[k0].z;
      double d = sqrt(dx*dx+dy*dy+dz*dz)+1;  // Add 1 to d to dampen movement
      double f = 1/(d*d*d);                  // Normalize and scale to 1/r^2
      a += f*dx;
      b += f*dy;
      c += f*dz;
   }
   //  Update velocity
   stars[k1].u = stars[k0].u + dt*a;
   stars[k1].v = stars[k0].v + dt*b;
   stars[k1].w = stars[k0].w + dt*c;
   //  Update position
   stars[k1].x = stars[k0].x + dt*stars[k1].u;
   stars[k1].y = stars[k0].y + dt*stars[k1].v;
   stars[k1].z = stars[k0].z + dt*stars[k1].w;
}

//
//  Advance time one time step
//
void Ex20opengl::Step()
{
   int k;
   //  Switch source and destination
   src = src?0:N;
   dst = dst?0:N;
   //  OpenMP
   if (mode)
      #pragma omp parallel for
      for (k=0;k<N;k++)
         Move(k);
   //  Sequential
   else
      for (k=0;k<N;k++)
         Move(k);
}

//
//  Scaled random value
//
void rand3(float Sx,float Sy,float Sz,float& X,float& Y,float& Z)
{
   float x = 0;
   float y = 0;
   float z = 0;
   float d = 2;
   while (d>1)
   {
      x = rand()/(0.5*RAND_MAX)-1;
      y = rand()/(0.5*RAND_MAX)-1;
      z = rand()/(0.5*RAND_MAX)-1;
      d = x*x+y*y+z*z;
   }
   X = Sx*x;
   Y = Sy*y;
   Z = Sz*z;
}

//
//  Initialize nBody problem
//
void Ex20opengl::InitLoc()
{
   //  Allocate room for twice as many bodies to facilitate ping-pong
   if (!stars) stars = (Star*)malloc(2*N*sizeof(Star));
   if (!stars) Fatal("Error allocating memory for stars\n");
   src = N;
   dst = 0;
   //  Assign random locations
   for (int k=0;k<N;k++)
   {
      rand3(dim/2,dim/2,dim/3,stars[k].x,stars[k].y,stars[k].z);
      rand3(vel,vel,vel,stars[k].u,stars[k].v,stars[k].w);
      switch (k%3)
      {
         case 0:
           stars[k].r = 1.0;
           stars[k].g = 1.0;
           stars[k].b = 1.0;
           break;
         case 1:
           stars[k].r = 1.0;
           stars[k].g = 0.9;
           stars[k].b = 0.5;
           break;
         case 2:
           stars[k].r = 0.5;
           stars[k].g = 0.9;
           stars[k].b = 1.0;
           break;
      }
      stars[k+N].r = stars[k].r;
      stars[k+N].g = stars[k].g;
      stars[k+N].b = stars[k].b;
   }
}

//
//  Constructor
//
Ex20opengl::Ex20opengl(QWidget* parent)
    : QGLWidget(parent)
{
   mode = 0;
   init  = false;
   mouse = false;
   asp = 1;
   dim = 10;
   vel = 0.1;
   fov = 55;
   th = ph = 0;
   x0 = y0 = 0;
   z0 = 1;
   zh = 0;
   stars = 0;
   sec0 = 0;
   fps = 0;
   N = 2048;
   InitLoc();
}

//
//  Reset view
//
void Ex20opengl::reset()
{
   th = ph = 0;
   dim = 10;
   Projection();
   //  Request redisplay
   updateGL();
}

//
//  Set shader
//
void Ex20opengl::setShader(int sel)
{
   mode = sel;
   //  Request redisplay
   updateGL();
}

//
//  Initialize
//
void Ex20opengl::initializeGL()
{
   if (init) return;
   init = true;

   Projection();

   //  Load shaders
   Shader(shader,":/ex20.vert",":/ex20.geom",":/ex20.frag");

   //  Load smoke particle
   QPixmap star(":/star.png");
   bindTexture(star,GL_TEXTURE_2D);

   //  Start 100 fps timer connected to updateGL
   move = true;
   timer.setInterval(1);
   connect(&timer,SIGNAL(timeout()),this,SLOT(updateGL()));
   timer.start();
   time.start();
}

//
//  Set projection when window is resized
//
void Ex20opengl::resizeGL(int width, int height)
{
   //  Window aspect ration
   asp = height ? width / (float)height : 1;
   //  Viewport is whole screen
   glViewport(0,0,width,height);
   //  Set projection
   Projection();
}

//
//  Draw the window
//
void Ex20opengl::paintGL()
{
   fps++;
   int sec = time.elapsed()/1000;
   if (sec!=sec0)
   {
      emit FPS(QString::number(fps));
      fps = 0;
      sec0 = sec;
   }
   //  Wall time (seconds)
   float t = 0.001*time.elapsed();
   if (move) zh = fmod(90*t,360);

   //  Clear screen and Z-buffer
   glClear(GL_COLOR_BUFFER_BIT);

   //  Set view
   glLoadIdentity();
   if (fov) glTranslated(0,0,-2*dim);
   glRotated(ph,1,0,0);
   glRotated(th,0,1,0);

   //  Integrate
   Step();

   //  Apply shader
   if (mode==2)
   {
      shader.bind();
      shader.setUniformValue("star",0);
      shader.setUniformValue("size",0.3f);
      glBlendFunc(GL_ONE,GL_ONE);
      glEnable(GL_BLEND);
   }

   //  Draw stars using vertex arrays
   glEnableClientState(GL_VERTEX_ARRAY);
   glEnableClientState(GL_COLOR_ARRAY);
   glVertexPointer(3,GL_FLOAT,sizeof(Star),&stars[0].x);
   glColorPointer(3,GL_FLOAT,sizeof(Star),&stars[0].r);
   //  Draw all stars from dst count N
   glDrawArrays(GL_POINTS,dst,N);
   //  Disable vertex arrays
   glDisableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_COLOR_ARRAY);

   //  Unset shader
   if (mode==2)
   {
      shader.release();
      glDisable(GL_BLEND);
   }

   //  Emit angles to display
   emit angles(QString::number(th)+","+QString::number(ph));
}

//
//  Throw a fatal error and die
//
void Ex20opengl::Fatal(QString message)
{
   QMessageBox::critical(this,"Ex20opengl",message);
   QApplication::quit();
}

//
//  Set OpenGL projection
//
void Ex20opengl::Projection()
{
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   if (fov)
   {
      float zmin = dim/4;
      float zmax = 4*dim;
      float ydim = zmin*tan(fov*M_PI/360);
      float xdim = ydim*asp;
      glFrustum(-xdim,+xdim,-ydim,+ydim,zmin,zmax);
   }
   else
      glOrtho(-dim*asp, +dim*asp, -dim, +dim, -dim, +dim);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

/******************************************************************/
/*************************  Mouse Events  *************************/
/******************************************************************/
//
//  Mouse pressed
//
void Ex20opengl::mousePressEvent(QMouseEvent* e)
{
   mouse = true;
   pos = e->pos();  //  Remember mouse location
}

//
//  Mouse released
//
void Ex20opengl::mouseReleaseEvent(QMouseEvent*)
{
    mouse = false;
}

//
//  Mouse moved
//
void Ex20opengl::mouseMoveEvent(QMouseEvent* e)
{
   if (mouse)
   {
      QPoint d = e->pos()-pos;  //  Change in mouse location
      th = (th+d.x())%360;      //  Translate x movement to azimuth
      ph = (ph+d.y())%360;      //  Translate y movement to elevation
      pos = e->pos();           //  Remember new location
      updateGL();               //  Request redisplay
   }
}

//
//  Mouse wheel
//
void Ex20opengl::wheelEvent(QWheelEvent* e)
{
   //  Zoom out
   if (e->delta()<0)
      dim += 0.1;
   //  Zoom in
   else if (dim>1)
      dim -= 0.1;
   //  Request redisplay
   Projection();
   updateGL();
}

//
//  Load shader
//
void Ex20opengl::Shader(QGLShaderProgram& shader,QString vert,QString geom,QString frag)
{
   //  Vertex shader
   if (vert.length() && !shader.addShaderFromSourceFile(QGLShader::Vertex,vert))
      Fatal("Error compiling "+vert+"\n"+shader.log());
   //  Geometry shader
   if (geom.length() && !shader.addShaderFromSourceFile(QGLShader::Geometry,geom))
      Fatal("Error compiling "+geom+"\n"+shader.log());
   //  Fragment shader
   if (frag.length() && !shader.addShaderFromSourceFile(QGLShader::Fragment,frag))
      Fatal("Error compiling "+frag+"\n"+shader.log());
   //  Link
   if (!shader.link())
      Fatal("Error linking shader\n"+shader.log());
}
